"""
Classes for describing asset transforms to storage backends.
"""

import copy

__all__ = [
    'BaseTransform',
    'CustomTransform'
]


class TransformMeta(type):
    """
    Meta class for transforms which adds the transform class to the register
    of transforms.
    """

    def __new__(meta, name, bases, dct):

        cls = super().__new__(meta, name, bases, dct)

        if name != 'BaseTransform' \
                and bases \
                and not cls._exclude_from_register:

            # Register the transform
            BaseTransform._transforms[cls._id] = cls

        return cls


class BaseTransform(metaclass=TransformMeta):
    """
    Transform classes provide a mechanism for describe how to transform an
    asset, for example rotating and cropping an image, converting it to a
    webp, etc.

    Each backend asset manager is responsible for translating a transform
    instance into the necessary instuctions for the storage service it
    represents.

    An transform is made up of a unique Id and the settings that will be used
    when applying it. Settings should be set as a dictionary and values should
    be JSON safe.

    An transform's Id is typically set as a static class property, and should
    use the following format:

        `{asset_type}.{transform}` -> `image.crop`

    backend (service) specific transforms should be postfixed, e.g:

        `{backend}.{asset_type}.{transform}` -> `hangar51.image.single_frame`

    All backends should support non-specific transforms.
    """

    # The transform's unique Id
    _id = ''

    # A table of all registered transforms (e.g `{id: transform_class}`)
    _transforms = {}

    # Flag indicating if the transform should be excluded from the register of
    # trasforms.
    _exclude_from_register = False

    def __init__(self, settings=None):

        # The settings that will be used when applying the transform
        if settings is None:
            settings = {}

        self._settings = settings

    @property
    def id(self):
        return self._id

    @property
    def settings(self):
        return copy.deepcopy(self._settings)

    def to_json_type(self):
        """
        Return a dictionary for the transform with values converted to JSON
        safe types.
        """
        return {'id': self.id, 'settings': self.settings}

    @classmethod
    def from_json_type(cls, json_type_data):
        """Convert a JSON safe dictionary of to an transform instance"""

        transform_cls = cls._transforms.get(json_type_data['id'])

        if transform_cls:
            return transform_cls(**json_type_data['settings'])

        return CustomTransform(
            json_type_data['id'],
            **json_type_data['settings']
        )


class CustomTransform(BaseTransform):
    """
    A custom transform, typically returned by the `from_json_type` class
    method when their is no matching transform registered (useful for ad hoc,
    backend / service specific transforms).
    """

    _exclude_from_register = True

    def __init__(self, id, **settings):

        # The Id for the custom transform
        self._id = id

        # The settings that will be used when applying the transform
        self._settings = settings



import json

import flask
from manhattan.forms import fields
from wtforms import widgets

from manhattan.assets.image_sets import ImageSet
from manhattan.assets.validators import AssetType

__all__ = [
    'AssetField',
    'ImageField',
    'ImageSetField'
]


class AssetField(fields.Field):
    """
    A field that supports a file being uploaded.
    """

    widget = widgets.HiddenInput()

    def __init__(self, label=None, validators=None, render_kw=None, **kwargs):

        # Ensure the field is flagged as an asset field
        if not render_kw:
            render_kw = {}

        render_kw['data-mh-file-field'] = True

        # Flag indicating if the asset's base transforms have changed (which
        # would indicate that the asset's variations need to be regenerated).
        self.base_transform_modified = False

        super().__init__(label, validators, render_kw=render_kw, **kwargs)

    def __call__(self, **kwargs):

        if flask.has_request_context() \
                and 'data-mh-file-field--blueprint' not in kwargs:

            # Add the current blueprint to the field arguments
            kwargs['data-mh-file-field--blueprint'] = flask.request.blueprint

        return super().__call__(**kwargs)

    def _value(self):
        if self.raw_data:
            return ' '.join(self.raw_data)
        elif self.data is not None:
            return json.dumps(self.data.to_json_type_for_field())
        return ''

    def process_data(self, value):
        """Process the the fields initial or default value"""
        self.data = value

    def process_formdata(self, values):
        """Process a value(s) submitted to the form"""
        if not values or not values[0]:
            self.data = None
            return

        # Convert the serialized asset JSON to an asset
        asset_data = json.loads(values[0])
        key = asset_data['key']
        base_transforms = asset_data.get('base_transforms')
        user_meta = asset_data.get('user_meta')

        if self.data and self.data.key == key:
            # Asset was original set as a value for the field
            asset = self.data

        else:
            # Attempt to retrieve asset from the temporary cache
            asset = flask.current_app.asset_mgr.get_temporary_by_key(key)

        if not asset:
            self.data = None
            return

        # Set any user defined meta data against the asset
        asset.user_meta = user_meta

        # For the base transforms we perform a comparison to flag that they
        # have been modified and therefore we need to regenerate existing
        # assets.
        old_transforms = json.dumps(asset.base_transforms)
        new_transforms = json.dumps(base_transforms)
        self.base_transform_modified = old_transforms != new_transforms

        asset.base_transforms = base_transforms

        # Cater for a preview URI provided client-side (required to provide
        # a preview after the field validates but the form fails to).
        if asset_data.get('preview_uri'):
            asset.preview_uri = asset_data['preview_uri']

        self.data = asset


class ImageField(AssetField):
    """
    A field that supports an image being uploaded.
    """

    def __init__(self,
        label=None,
        validators=None,
        render_kw=None,
        crop_aspect_ratio=None,
        fix_crop_aspect_ratio=False,
        **kwargs
    ):

        if not validators:
            validators = []

        # Require the asset be an image
        validators.append(AssetType('image'))

        if not render_kw:
            render_kw = {}

        # Flag the file type for the fields as an image
        render_kw['data-mh-file-field--file-type'] = 'image'

        # Set the label for the field (it not already set)
        render_kw['data-mh-file-field--label'] = render_kw.pop(
            'data-mh-file-field--label',
            'Upload an image...'
        )

        # Add crop settings to the render keywords
        if crop_aspect_ratio != None:
           render_kw['data-mh-file-field--crop-aspect-ratio'] \
                = crop_aspect_ratio

        if fix_crop_aspect_ratio:
            render_kw['data-mh-file-field--fix-crop-aspect-ratio'] = True

        # Default to accepting only images
        if 'data-mh-file-field--accept' not in render_kw:
            render_kw['data-mh-file-field--accept'] = 'image/*'

        super().__init__(label, validators, render_kw=render_kw, **kwargs)


class ImageSetField(fields.Field):
    """
    A field that supports a set of images being uploaded.

    Usage:

        class MyForm(BaseForm):

            my_image = ImageSetField(
                'My image',
                versions=[
                    ('l', 'Desktop', 2.0), # First is set as the base version
                    ('m', 'Tablet', 1.5),
                    ('s', 'Mobile', 1.0)
                ],
                validators=[validators.Required()]
            )

    Note: Each version is given as a tuple of the form:

    - version name
    - version label
    - crop aspect ratio for the version (optional)

    """

    widget = widgets.HiddenInput()

    def __init__(self,
        label=None,
        versions=None,
        fix_crop_aspect_ratio=False,
        validators=None,
        render_kw=None,
        **kwargs
    ):

        assert versions, 'You must give at least one version for an image set.'

        version_length = len(versions[0])
        assert version_length in [2, 3], (
            '`versions` argument must me given as '
            '(name, label) or (name, label, crop_ratio).'
        )
        for version in versions:
            assert len(version) == version_length, (
                'Each item in the `versions` argument must be the same '
                'length.'
            )

        assert not fix_crop_aspect_ratio or version_length == 3, (
            '`fix_crop_aspect_ratio` can not be true unless crop ratios have '
            'been given for each version.'
        )

        if not render_kw:
            render_kw = {}

        # Store the versions the image set supports
        self.versions = versions

        # Set the versions for the field
        render_kw['data-mh-image-set--versions'] = ','.join([
            v[0] for v in versions
        ])

        # Set the labels for each version in the image set
        render_kw['data-mh-image-set--version-labels'] = ','.join([
            v[1] for v in versions
        ])

        # Set the crop ratio for each version in the image set
        if version_length == 3:
            render_kw['data-mh-image-set--crop-aspect-ratios'] = ','.join([
                str(v[2]) for v in versions
            ])

        if fix_crop_aspect_ratio:
            render_kw['data-mh-image-set--fix-crop-aspect-ratio'] = True

        # Default to accepting only images
        if 'data-mh-image-set--accept' not in render_kw:
            render_kw['data-mh-image-set--accept'] = 'image/*'

        # Ensure the field is flagged as an image set field
        render_kw['data-mh-image-set'] = True

        # A set of keys indicating which of the image set's base transforms
        # have changed (which would indicate that the associated asset
        # variations need to be regenerated).
        self.modified_base_transforms = set([])

        super().__init__(label, validators, render_kw=render_kw, **kwargs)

    def __call__(self, **kwargs):

        if flask.has_request_context() \
                and 'data-mh-image-set--blueprint' not in kwargs:

            # Add the current blueprint to the field arguments
            kwargs['data-mh-image-set--blueprint'] \
                    = flask.request.blueprint

        return super().__call__(**kwargs)

    def _value(self):
        if self.raw_data:
            return ' '.join(self.raw_data)
        elif self.data is not None:
            return json.dumps(self.data.to_json_type_for_field())
        return ''

    def process_data(self, value):
        """Process the the fields initial or default value"""
        self.data = value

    def process_formdata(self, values):
        """Process a value(s) submitted to the form"""
        if not values or not values[0]:
            self.data = None
            return

        # Convert the serialized JSON to an image set
        image_set_data = json.loads(values[0])
        images = image_set_data.get('images', {})
        base_transforms = image_set_data.get('base_transforms', {})
        base_version = image_set_data.get('base_version', self.versions[0][0])
        alt = image_set_data.get('alt', '')

        # Build a list of valid versions
        valid_versions = [v[0] for v in self.versions]

        # Check the base version is one of the valid versions
        if base_version not in valid_versions:
            self.data = None
            return

        # Check that all base transforms are valid
        for version, transforms in base_transforms.items():
            if version not in valid_versions:
                self.data = None
                return

        # (Re)build the image set instance ensuring data is trusted
        image_set = ImageSet(
            images={},
            base_transforms=base_transforms,
            base_version=base_version,
            alt=alt
        )

        for version in valid_versions:

            # Convert the assets within the image set to `Asset` instances
            image_data = image_set_data['images'].get(version)

            if image_data:
                image = None
                key = image_data.get('key')

                if (
                    self.data
                    and version in self.data.images
                    and self.data.images[version].key == key
                ):
                    # Asset initial set against value for the field
                    image = self.data.images[version]

                else:
                    # Attempt to retrieve asset from the temporary cache
                    image = flask\
                            .current_app.asset_mgr.get_temporary_by_key(key)

                if image and image.type == 'image':
                    image_set.images[version] = image

            # Cater for a preview URIs provided client-side (required to
            # provide a preview after the field validates but the form fails
            # to).
            if image_set_data.get('preview_uris'):
                image_set.preview_uris = image_set_data['preview_uris']

            # Determine if the user has changed the base transform for the
            # version.
            old_transforms = None
            new_transforms = base_transforms.get(version)

            if self.data:
                old_transforms = self.data.base_transforms.get(version)

            if json.dumps(old_transforms) != json.dumps(new_transforms):
                self.modified_base_transforms.add(version)

        # Set the data for the field
        self.data = None
        if image_set.base_image:
            self.data = image_set




"""
Classes for implementing a backend for asset management.
"""

import json

from manhattan.assets import Asset

__all__ = ['BaseAssetMgr']


# Classes

class BaseAssetMgr:
    """
    Assets are managed via an asset manager class which must provide the API
    defined by the `BaseAssetMgr` class.
    """

    # A table of functions that convert analyzers to a format understood by
    # the backend service.
    _analyzer_converters = None

    # A table of functions that convert transforms to a format understood by
    # the backend service.
    _transform_converters = None

    def __init__(self, cache, expire_after=3600):
        # The cache used to store references to temporary assets, must support
        # `get` and `set` methods, `set` must support an `timeout` argument
        # when the data will expire.
        #
        # `werkzeug.contrib.cache` provides a number of suitable classes for
        # creating caches (e.g `SimpleCache` and `MemcachedCache`).
        self._cache = cache

        # The amount of time before temporary assets expire
        self._expire_after = expire_after

    def analyze(self, asset, analyzers):
        """Analyze the asset"""
        raise NotImplementedError()

    def analyze_many(self, assets, analyzers):
        """Analyze the given list of asseta"""
        for asset in assets:
            self.generate_variations(asset, analyzers[asset.key])

    def clear_cache(self, asset):
        """Clear the given asset from the cache"""
        self._cache.delete(self.get_cache_key(asset.key))

    def clone(self, asset, name=None, secure=None):
        """Returning a temporary asset cloned from the specified asset"""
        secure = asset.secure if secure is None else secure

        # Retrieve the asset we're cloning
        file = self.retrieve(asset)

        # Store the file as a new temporary asset (cloning it)
        cloned = self.store_temporary(
            (file, asset.filename),
            name=name,
            secure=secure
        )

        return cloned

    def generate_variations(self, asset, variations, base_transforms=None):
        """Generate variations for the asset"""
        raise NotImplementedError()

    def generate_variations_for_many(
        self,
        assets,
        variations,
        base_transforms=None
    ):
        """Generate variations for the given assets"""
        base_transforms = base_transforms or {}

        for asset in assets:
            self.generate_variations(
                asset,
                variations[asset.key],
                base_transforms.get(asset.key, None)
            )

    def get_temporary_by_key(self, key):
        """Get a temporary asset by it's key"""
        asset_json = self._cache.get(self.get_cache_key(key))
        if asset_json:
            asset = Asset(json.loads(asset_json))
            for name, variation in asset.variations.items():
                asset.variations[name] = Asset(variation)
            return asset

    def persist_many(self, assets):
        """Set the given list of assets as permenant"""
        for asset in assets:
            self.store(asset)

    def remove(self, asset):
        """Remove the specified asset"""
        raise NotImplementedError()

    def remove_many(self, assets):
        """Remove the given list of assets"""
        for asset in assets:
            self.remove(asset)

    def retrieve(self, asset):
        """Retrieve the asset"""
        raise NotImplementedError()

    def store_temporary(self, file, name=None, secure=False):
        """Store an asset temporarily"""
        raise NotImplementedError()

    def store(self, file_or_asset, name=None, secure=False):
        """
        Store an asset.

        NOTE: The `name` argument is ignored if an asset is provided, to rename
        an existing asset you must clone the asset with a new name and then
        store the resulting temporary asset.

        NOTE: The `secure` argument is ignored if an asset is provided.
        """
        raise NotImplementedError()

    def update_cache(self, asset):
        """Update the given asset in the cache"""
        if asset.temporary:
            self._cache.set(
                self.get_cache_key(asset.key),
                json.dumps(asset.to_json_type()),
                self._expire_after
            )

    # Class methods

    @classmethod
    def get_cache_key(cls, asset_key):
        return f'temporary_assets:{asset_key}'

    @classmethod
    def convert_analyzers(cls, analyzers):
        """
        Convert a stack of analyzers to a format understood by the backend
        service.
        """

        backend_analyzers = []
        for analyzer in analyzers:
            converter = cls._analyzer_converters.get(analyzer.id)
            if not converter:
                converter = cls._analyzer_converters.get('__custom__')
            backend_analyzers.append(converter(analyzer))

        return backend_analyzers

    @classmethod
    def convert_transforms(cls, transforms):
        """
        Convert a stack of transforms to a format understood by the backend
        service.
        """

        backend_transforms = []
        for transform in transforms:
            converter = cls._transform_converters.get(transform.id)
            if not converter:
                converter = cls._transform_converters.get('__custom__')
            backend_transforms.append(converter(transform))

        return backend_transforms




from mongoframes import SubFrame

from .assets import Asset

__all__ = ['ImageSet']


class ImageSet(SubFrame):
    """
    `ImageSets` are used specificially to implement <picture> tags in HTML.
    ImageSets provide (typically) multiple version of an image to be displayed for
    different media queries. Each image may be either a seperate transform
    (crop/rotate) of a base image (asset) or can be a unique image (asset).

    Usage (building a picture element in a template):

        <picture class="article__image">
            {% for srcset, media in image_set.sources(
                l='(min-width: 960px)',
                m='(min-width: 720px)'
            ) %}
                {{ source|safe }}
            {% endif %}

            <img
                src="{{ image_set.url('s', 'm', 'l') }}"
                alt="{{ image_set.alt }}"
            >
        </picture>

    """

    _fields = {
        # A table of images that represent the image set, if a key for an
        # image is not present in the table then we reference the base image.
        'images',

        # A table of transforms (crop/rotate) each image within the image set,
        # if a key for an image is not present in the table then we reference
        # the transform for the base image, and if no base image transforms are
        # present then the base image us used.
        'base_transforms',

        # The verion of the image within the `images` table that represents
        # the base image. The first image uploaded for an image set is
        # generally considered the base image, other images in the image set
        # will often be different crops of the base image.
        'base_version',

        # An alt tag for the image represented by the image set
        'alt'
    }

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

        # A map of preview URIs for the assets set when assets are been
        # modified client-side. This value is never stored in the database and
        # is used to allow a preview to persist between multiple form
        # submissions.
        self.preview_uris = None

    @property
    def base_image(self):
        return self.images.get(self.base_version)

    @property
    def versions(self):
        versions = []
        for image in self.images.values():
            for version in image.variations.keys():
                versions.append(version)
        return list(set(versions))

    def variation(self, version):
        """Return the image for the given version"""
        image = self.images.get(version, self.base_image)
        if version in image.variations:
            return image.variations[version]

    def sources(self, **versions):
        """
        Return a list of srcset URLs and media queries for the given versions.

        Usage:

            versions = {
                l='(min-width: 960px)',
                m='(min-width: 720px)'
            }

            for url, media in image_set.sources(versions):
                ...
        """
        sources = []

        base_image = self.base_image
        for version, media in versions.items():
            image = self.images.get(version, base_image)

            if version in image.variations:
                sources.append((image.variations[version].url, media))

        return sources

    def to_json_type_for_field(self):
        document_dict = self._json_safe_for_field(self._document)
        self._remove_keys(document_dict, self._private_fields)

        if self.preview_uris:
            document_dict['preview_uris'] = self.preview_uris

        return document_dict

    def url(self, version):
        """Return the URL for the given version"""
        variation = self.variation(version)
        if variation:
            return variation.url

    @classmethod
    def _json_safe_for_field(cls, value):
        """Return a JSON safe value"""

        if isinstance(value, Asset):
            return value.to_json_type_for_field()

        # Lists
        elif isinstance(value, (list, tuple)):
            return [cls._json_safe_for_field(v) for v in value]

        # Dictionaries
        elif isinstance(value, dict):
            return {
                k: cls._json_safe_for_field(v)
                for k, v in value.items()
            }

        return cls._json_safe(value)

    @classmethod
    def default_projection(cls):
        """
        Short-cut for defining a simple projection, usage:

            projection = {
                'image_set': ImageSet.default_projection()
            }

        """
        projection = {
            '$sub': cls,
            'images': {
                '$sub.': Asset,
                'variations': {'$sub.': Asset}
            }
        }
        return projection



"""
Generic upload asset chain.

NOTE: The generic upload chain is designed to handle a single file upload,
multiple file uploads client side should result in multiple calls to the upload
chain.
"""

import os

import flask
from manhattan.assets.transforms.images import Fit, Output
from manhattan.assets.backends import exceptions
from manhattan.chains import Chain, ChainMgr

from manhattan.manage.views import factories
from manhattan.manage.views import utils
from manhattan.manage.utils import get_config

__all__ = ['upload_chains']


# Define the chains
upload_chains = ChainMgr()

# POST
upload_chains['post'] = Chain([
    'config',
    'authenticate',
    'store_temporary_asset',
    'render_json'
])


# Define the links
upload_chains.set_link(factories.config())
upload_chains.set_link(factories.authenticate())

@upload_chains.link
def store_temporary_asset(state):
    """
    Store the file uploaded as a temporary asset.

    The client request is expected to include a file under the parameter name of
    `file`.

    If the file is successfully stored this links adds an `asset` key to the
    state containing the `Asset` instance representing the uploaded file.

    IMPORTANT: Validation of the asset itself is the responsibility of the view
    that is storing the associated parent document, not the upload chain. The
    upload chain merely validates that a file was provided and could be stored.
    """

    # Check a file was provided
    if 'file' not in flask.request.files \
            or not flask.request.files['file'].filename:
        return utils.json_fail('No file sent')

    # Apply the whitelist to
    file = flask.request.files['file']

    if state.ext_whitelist is None:
        state.ext_whitelist = [
            '.7z',
            '.csv',
            '.docx',
            '.eml',
            '.gif',
            '.jpeg',
            '.jpg',
            '.key',
            '.keynote',
            '.md',
            '.mp3',
            '.msg',
            '.odf',
            '.ods',
            '.odt',
            '.ogg',
            '.pages',
            '.pdf',
            '.png',
            '.ppt',
            '.pptx',
            '.rar',
            '.rtf',
            '.svg',
            '.tif',
            '.tiff',
            '.txt',
            '.webp',
            '.xls',
            '.xlsx',
            '.zip'
        ]

    if len(state.ext_whitelist):

        filename, ext = os.path.splitext(file.filename)

        if ext.lower() not in state.ext_whitelist:
            return utils.json_fail('Not a supported file type.')

    # Attempt to store the file
    asset_mgr = flask.current_app.asset_mgr
    asset_name = state.asset_name

    if callable(asset_name):
        asset_name = asset_name(file.filename)

    try:

        state.asset = asset_mgr.store_temporary(
            (file.filename, file),
            name=asset_name,
            secure=flask.request.values.get('secure') == 'secure'
        )

    except exceptions.StoreError as e:
        return utils.json_fail(str(e))

    # Check to see if an image file type has been requested and if so validate
    # the asset backend recognized it as an image.
    file_type = flask.request.values.get('file_type')
    if file_type == 'image':
        if state.asset.type != file_type:
            return utils.json_fail('Not a supported: ' + file_type)

    # Attempt to get the config manager associated with the upload
    blueprint_name = flask.request.values.get('blueprint', None)
    blueprint_config = None
    if blueprint_name:
        blueprint_config = get_config(flask.current_app, blueprint_name)

    # Look up any analyzers and variations (transforms) associated with the
    # uploaded file.
    field_name = flask.request.values.get('field_name', None)

    analyzers = {}
    variations = {}

    if blueprint_config and field_name:
        analyzers = blueprint_config.asset_analyzers.get(field_name, {})
        variations = blueprint_config.asset_variations.get(field_name, {})

        # Image sets may request that only one specific variations is
        # generated (e.g non-base versions only generate their respective
        # variation).
        version = flask.request.values.get('version')
        if version and version in variations:
            variations = {version: variations[version]}

    # We perform any analysis and transformation of assets eagerly (e.g when
    # they are temporary). Whilst variations may need to be regenerated for an
    # asset before storing it permanently (if it has base transforms) for the
    # majority of cases eager generation distributes the work load better.

    if analyzers:

        # Perform analysis
        try:
            asset_mgr.analyze(state.asset, analyzers)

        except exceptions.StoreError as e:
            return utils.json_fail(str(e))

    in_page_upload = flask.request.values.get('in_page') == 'in_page'
    if variations or in_page_upload:

        if state.asset.type == 'image':

            # Add system variations for the image (required for the thumbnail
            # preview and draft that appears in the image editor).

            # Thumbnail
            if variations:
                variations['--thumb--'] = flask.current_app.config.get(
                    'ASSET_THUMB_VARIATION',
                    [
                        Fit(480),
                        Output('jpg', 75)
                    ]
                )

            # Draft
            variations['--draft--'] = flask.current_app.config.get(
                'ASSET_DRAFT_VARIATION',
                [
                    Fit(1200),
                    Output('jpg', 50)
                ]
            )

        # Perform transforms
        try:
            asset_mgr.generate_variations(state.asset, variations)

        except exceptions.StoreError as e:
            return utils.json_fail(str(e))

        if '--thumb--' in variations:

            # Store the thumbnail transforms so the variation can be
            # regenerated outside of this view (draft variations are not
            # regenerated as a rule).
            state.asset.variations['--thumb--'].local_transforms = [
                t.to_json_type() for t in variations['--thumb--']
            ]

    # Store a temporary copy of the temporary asset
    asset_mgr.update_cache(state.asset)

@upload_chains.link
def render_json(state):
    """
    Return a successful response with the uploaded `asset` included in the
    payload.
    """
    return utils.json_success({'asset': state.asset.to_json_type_for_field()})




from manhattan.assets.transforms.base import BaseTransform

__all__ = ['Output']


class Output(BaseTransform):
    """
    Set the output format of an image. The list of support formats will vary
    depending on service, however as a minimum the following formats should be
    supported by all services:

    - jpg
    - gif
    - png
    - webp
    """

    _id = 'image.output'

    def __init__(self, format, quality=None, lossless=None):
        settings = {'format': format}
        if quality is not None:
            settings['quality'] = quality

        if lossless is not None:
            settings['lossless'] = lossless

        super().__init__(settings)
